/*-------------------------------------------------------------------*/
/* CBMODIFY.C   Pasi Frnti                                          */
/*              Timo Kaukoranta                                      */
/*              Juha Kivijrvi                                       */
/*              Marko Tuononen                                       */
/*                                                                   */
/* Modifies an existing codebook:                                    */
/*  - Sorting                                                        */
/*  - Rescaling                                                      */
/*  - Changing dimensions                                            */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName      "CBMODIFY"
#define VersionNumber "Version 0.21"  /* MT: made immune to 
                                         erroneous TrainingSetID */ 
#define LastUpdated   "12.05.05"

/* ------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define  FACTFILE "cbmodify.fac"
#include "cb.h"
#include "interfc.h"
#include "file.h"
#include "memctrl.h"
#include "parametr.c"
#include "random.h"
#include "sortcb.h"

#define MAXFILENAME      64

#ifndef WIN32
  #define stricmp strcasecmp
  #define strnicmp strncasecmp
#endif


/*============================  P R I N T  ===============================*/


void PrintInfo(void)
{
  printf("%s     %s                (last modified %s)\n",ProgName,VersionNumber,LastUpdated);
  printf("\nThis program modifies existing codebook (CB-file).\n");
  printf("Duplicates will also be combined in the resulting codebook.\n");
  printf("Usage: %s inputfile [outputfile] [%coptions]\n", ProgName, OPTION_SYMBOL);
  printf("\n  Options:\n");
  PrintOptions();
  printf("%s     %s\n",ProgName,VersionNumber);
}


/*-------------------------------------------------------------------*/


static void PrintOperatingInfo(char* InputFileName,
                               char* OutputFileName)
{
  if( Value(QuietLevel) >= 1 )
    {
    printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
    printf("Input file:       %s\n", InputFileName);
    printf("Output file:      %s\n", OutputFileName);
    printf("\n");
    PrintSelectedOptions();
    printf("\n");
    }
}


/*-------------------------------------------------------------------*/


static void PrintProgress(int processed, int total)
{
  switch(  Value(QuietLevel) )
     {
     case 0:  return;
     case 1:  return;
     default: printf("%5.1f %% done.\r", (float)(processed) * 100 / total);
              break;
     }
  fflush(stdout);
}


/* ==============  P A R A M E T E R   P A R S I N G  ============== */

CBFILETYPE DetermineFileType(char *FileName) 
{
  CBFILETYPE filetype;
  char* c;

  filetype = DetermineCBFileType(FileName);
  if (filetype == NOTFOUND) 
    {
    c = strrchr(FileName, '.');
    if (c != NULL) 
      {
      /* determine type by extension */
      c++;
      if (stricmp(c, FormatNameCB)==0)
        filetype = CBFILE;
      else if (stricmp(c, FormatNameTS)==0)
        filetype = TSFILE;
      }

    if (filetype == NOTFOUND)
      {
      printf("Cannot determine type of input file (%s).\n", FileName);
      exit(-1);            
      }        
    }
    return filetype;
}

static CBFILETYPE AddExtensions(char*  FileName1,
                          char*  FileName2)
{
  CBFILETYPE filetype;

  switch( Value(ChangeFileType) )
    {
    case CHANGETOTS:   
      filetype = TSFILE; 
      break;

    case CHANGETOCB:   
      filetype = CBFILE; 
      break;
    
    case SAMEFILETYPE: 
    default:
      filetype = DetermineFileType(FileName1); 
      break;
    }

  if (strlen(FileName2) < 1)
    {
    PickFileName(FileName1, FileName2);
    }
  CheckFileName(FileName2, FormatName(filetype));
  if( ! stricmp(FileName1,FileName2) )
    {
    printf("The names of input and output files must be different (%s).\n", FileName1);
    exit(-1);
    }
  return filetype;
}


/*=====================  Remove duplicates  =========================*/


int FindDuplicate(CODEBOOK* CB, int v)
/* Finds a duplicate vector from v+1..size-1 from the codebook.
   If no duplicate found, value v is returned. */
{
  int i;

  for(i=v+1; i<BookSize(CB); i++)
     {
     if(EqualVectors( Vector(CB,v), Vector(CB,i), VectorSize(CB) ))
		{
		return(i);
		}
	 }
  return(v);
}


/*-------------------------------------------------------------------*/


void CombineVectors(CODEBOOK* CB, int v1, int v2)
{
/*    char s[82]; */
   int  i;

   if(v1==v2 || v1<0 || v1>BookSize(CB) || v2<0 || v2>BookSize(CB)) return;
   if(v1>v2)  { i=v1; v1=v2; v2=i; }
   VectorFreq(CB,v1) = VectorFreq(CB,v1) + VectorFreq(CB,v2);
/* strcpy(s, VectorName(CB,v1));
   strncat(s, "+", 80);
   strncat(s, VectorName(CB,v2), 80);
   ChangeVectorName(&Node(CB,v1), s); */
   for(i=v2+1; i<BookSize(CB); i++)
      {
      CopyNode( &Node(CB,i), &Node(CB,i-1), VectorSize(CB) );
      }
   BookSize(CB)--;
}


/*-------------------------------------------------------------------*/


void RemoveDuplicates(CODEBOOK* CB)
{
  int  d, i=0;

  while(i<BookSize(CB)-1)
     {
     d = FindDuplicate(CB,i);
     if(d!=i)
        {
        CombineVectors(CB,i,d);
        }
     else
        {
        i++;
        PrintProgress(i, BookSize(CB));
        }
	 }
  PrintProgress(100,100);
}


/*========================  M O D I F Y  ============================*/


void PrepareNewCodebook(CODEBOOK* CB1, CODEBOOK* CB2, CBFILETYPE filetype)
{
  CBFILETYPE ft;

  switch( Value(ChangeFileType) )
    {
    case SAMEFILETYPE:/*ft = RecognizeCBFileType(CB1);*/
                       ft = filetype; break;
    case CHANGETOTS:   ft = TSFILE; break;
    case CHANGETOCB:   ft = CBFILE; break;
    default:           ft = RecognizeCBFileType(CB1); break;
    }
  CreateNewCodebook(CB2, BookSize(CB1), CB1);
  SetVersionString(CB2, ft);
  AddGenerationMethod(CB2, CB1->GenerationMethod);
  if(Value(Rescaling)==YES)
    {
    CB2->MaxValue = Value(NewMaximum);
    if (CB2->MaxValue > 255)
      CB2->BytesPerElement = 2;
    if (CB2->MaxValue > 65535)
      CB2->BytesPerElement = 4;
    }
  if( Value(ChangeDimensions)==YES )
    {
    if( VectorSize(CB1) < Value(Dimensions) )
      {
      ErrorMessage("ERROR: Original vector dimension (%i) is smaller than requested (%i)\n",
                    VectorSize(CB1), Value(Dimensions));
      ExitProcessing( -1 );
      }
    else
      {
      CB2->BlockSizeX = 1;
      CB2->BlockSizeY = Value(Dimensions);
      }
    }
}


/*-------------------------------------------------------------------*/


void CopyPartialVector( VECTORTYPE v1, VECTORTYPE v2, int Vsize1, int Vsize2)
{
  int i;

  for(i=0; i<Vsize2; i++)
	{
	v2[i] = (i<Vsize1) ? v1[i] : 0;
	}
}


/*-------------------------------------------------------------------*/


void CopyBookPartially(CODEBOOK* CB1, CODEBOOK* CB2)
{
  int i;

  for(i=0; i<BookSize(CB1); i++)
	{
    CopyNode( &Node(CB1,i), &Node(CB2,i), VectorSize(CB2) );
	}
}


/*-------------------------------------------------------------------*/


void ModifyVector(VECTORTYPE v, int Vsize, int OldMax)
{
  int   i;

  for(i=0; i<Vsize; i++)
     {
     v[i] = (int) (0.5 + (float) Value(NewMaximum) * v[i] / OldMax);
     }
}


/*-------------------------------------------------------------------*/


void AddNoiseToVector(CODEBOOK* CB, int v)
{
  int    i;
  double r;

  for(i=0; i<VectorSize(CB); i++)
     {
	 r = CB->MaxValue * dgauss() * Value(AddNoise) / 100.0;
     VectorScalar(CB,v,i) = (int) (VectorScalar(CB,v,i) + r + 0.5);
     if( VectorScalar(CB,v,i) < CB->MinValue )
         VectorScalar(CB,v,i) = CB->MinValue;
     if( VectorScalar(CB,v,i) > CB->MaxValue )
         VectorScalar(CB,v,i) = CB->MaxValue;
     }
}


/*-------------------------------------------------------------------*/


void ModifyCodebook(CODEBOOK* CB, int OldMax)
{
  int i;

  if( Value(QuietLevel)>=2 ) printf("Modifying codebook:\n");
  for(i=0; i<BookSize(CB); i++)
    {
    if(Value(Rescaling)==YES)
       {
       ModifyVector(Vector(CB,i), VectorSize(CB), OldMax);
       }
    if(Value(FrequencyOne)==YES)
       {
       VectorFreq(CB,i) = 1;
       }
    if( Value(AddNoise) > 0 )
       {
       AddNoiseToVector(CB,i);
       }
    PrintProgress(i+1, BookSize(CB));
    }
}


/*============================  MAIN  ===============================*/


int main(int argc, char* argv[])
{
  char         InputFileName[MAXFILENAME];
  char         OutputFileName[MAXFILENAME];
  CBFILETYPE   filetype;
  CODEBOOK     CB1, CB2;
  ParameterInfo paraminfo[2] =  { { InputFileName,  "", 0, INFILE  }, 
                                  { OutputFileName, "", 1, OUTFILE } };

  ParseParameters( argc, argv, 2, paraminfo);
  filetype = AddExtensions(InputFileName, OutputFileName);
  PrintOperatingInfo(InputFileName, OutputFileName);

  ReadCodebook(InputFileName, &CB1);
  PrepareNewCodebook(&CB1, &CB2, filetype);
  CopyBookPartially(&CB1, &CB2);
  ModifyCodebook(&CB2, CB1.MaxValue);

  if( Value(KeepDuplicates)==NO )
    {
    if( Value(QuietLevel)>=2 ) printf("Removing duplicates:\n");
    RemoveDuplicates(&CB2);
    }
  if( Value(Sorting) )
    {
    if( Value(QuietLevel)>=2 ) printf("Sorting codebook: ");
    SortCodebook(&CB2, Value(Sorting));
    if( Value(QuietLevel)>=2 ) printf("OK.\n");
    }

  WriteCodebook(OutputFileName, &CB2, Value(OverWrite));
  printf("\n");
  FreeCodebook(&CB1);
  FreeCodebook(&CB2);
  checkmemory();
  
  return 0;
}

